/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

package uk.co.petertribble.jangle;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Date;
import java.awt.event.*;
import java.math.BigInteger;

import org.jfree.chart.*;
import org.jfree.data.*;
import org.jfree.data.time.*;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.XYPlot;

public class SnmpChart implements ActionListener {

    private SnmpController sc;
    private JFreeChart chart;
    protected javax.swing.Timer timer;
    protected int delay = 30000;
    protected int maxage = 1800000;
    private TimeSeriesCollection dataset;
    private boolean showdelta;
    // times in milliseonds
    private long lastsnap;
    private long newsnap;
    private String charttitle;

    private Map <String, TimeSeries> tsmap;
    // save previous values for rates
    private Map <String, BigInteger> valueMap;

    public SnmpChart(SnmpController sc, String statistic) {
	this(sc, statistic, true);
    }

    public SnmpChart(SnmpController sc, String statistic,
		boolean showdelta) {
	this.sc = sc;
	this.showdelta = showdelta;
	charttitle = statistic;
	initialize(statistic);
    }

    public SnmpChart(SnmpController sc, List <String> statistics,
		boolean showdelta) {
	this.sc = sc;
	this.showdelta = showdelta;
	charttitle = "";
	initialize(statistics);
    }

    private void initialize(String statistic) {
	List <String> statistics = new ArrayList <String> ();
	statistics.add(statistic);
	initialize(statistics);
    }

    private void initialize(List <String> statistics) {
	tsmap = new HashMap <String, TimeSeries> ();
	valueMap = new HashMap <String, BigInteger> ();
	dataset = new TimeSeriesCollection();
	lastsnap = 0;

	for (String statistic : statistics) {
	    TimeSeries ts = new TimeSeries(statistic, Millisecond.class);
	    ts.setMaximumItemAge(maxage);
	    dataset.addSeries(ts);
	    tsmap.put(statistic, ts);
	}

	/*
	 * Initialize the saved values so that the first calculated rate
	 * will be the average since statistic creation.
	 */
	for (String statistic : statistics) {
	    valueMap.put(statistic, BigInteger.ZERO);
	}

	updateAccessory();

	String ylabel = showdelta ? SnmpResources.getString("CHART.RATE")
	    : SnmpResources.getString("CHART.VALUE");

	chart = ChartFactory.createTimeSeriesChart(
		charttitle,
		SnmpResources.getString("CHART.TIME"),
		ylabel,
		dataset,
		true,
		true,
		false);

	XYPlot xyplot = chart.getXYPlot();
	NumberAxis loadaxis = new NumberAxis(ylabel);
	loadaxis.setAutoRange(true);
	loadaxis.setAutoRangeIncludesZero(true);
	xyplot.setRangeAxis(loadaxis);

	startLoop();
    }

    public JFreeChart getChart() {
	return chart;
    }

    /**
     * Add another statistic to the current chart. If the statistic is
     * already being charted, it is not added.
     */
    public void addStatistic(String statistic) {
	if (!tsmap.containsKey(statistic)) {
	    TimeSeries ts = new TimeSeries(statistic, Millisecond.class);
	    ts.setMaximumItemAge(maxage);
	    dataset.addSeries(ts);
	    tsmap.put(statistic, ts);
	    /*
	     * Save the current value, not zero. This ensures that the first
	     * rate calculation gives sensible values. It will be low, as it
	     * will average the change between now and the next update over
	     * the whole update interval. This is better than nothing, and
	     * the way to do this properly would be to have separate lastsnap
	     * values saved for each statistic. This would be overkill when
	     * the only affected datapoint is the first one.
	     */
	    // FIXME get the actual value
	    valueMap.put(statistic, BigInteger.ZERO);
	}
    }

    /**
     * Remove the requested statistic from the current chart.
     */
    // FIXME is it valid to remove all statistics?
    public void removeStatistic(String statistic) {
	if (tsmap.containsKey(statistic)) {
	    dataset.removeSeries(tsmap.get(statistic));
	    tsmap.remove(statistic);
	    valueMap.remove(statistic);
	}
    }

    /**
     * Set the maximum age of the chart. Only statistics younger than this
     * age will be shown.
     *
     * @param age The required maximum age in seconds.
     */
    public void setMaxAge(int age) {
	maxage = age*1000;
	// FIXME iterate over the values in the map
	for (String statistic : tsmap.keySet()) {
	    TimeSeries ts = tsmap.get(statistic);
	    ts.setMaximumItemAge(maxage);
	}
    }

    /**
     * Update the statistics
     */
    public void updateAccessory() {
	double value;
	Millisecond ms = new Millisecond();
	Date newTime = new Date();
	newsnap = newTime.getTime();
	// loop over all statistics
	for (String statistic : tsmap.keySet()) {
	    TimeSeries ts = tsmap.get(statistic);
	    if (showdelta) {
		BigInteger lastvalue = valueMap.get(statistic);
		SnmpObject sno = sc.getValue(statistic);
		BigInteger newvalue = sno.getNumber();
		BigInteger bd = newvalue.subtract(lastvalue);
		double d = bd.doubleValue();
		double dt = (double) (newsnap - lastsnap);
		value = 1000.0*(d/dt);
		valueMap.put(statistic, newvalue);
	    } else {
		SnmpObject sno = sc.getValue(statistic);
		BigInteger newvalue = sno.getNumber();
		value = newvalue.doubleValue();
	    }
	    ts.add(ms, value);
	}
	lastsnap = newsnap;
    }

    public void startLoop() {
	if (timer == null) {
	    timer = new javax.swing.Timer(delay, this);
	}
	timer.start();
    }

    public void stopLoop() {
	if (timer != null) {
	    timer.stop();
	}
    }

    public void setDelay(int interval) {
	delay = interval*1000;
	if (timer != null) {
	    timer.setDelay(delay);
	}
    }

    public void actionPerformed(ActionEvent e) {
	updateAccessory();
    }
}
